import hashlib
import os

from sanic.request import Request
from sanic.response import HTTPResponse, redirect

from sanic_security.core.config import config


def xss_prevention_middleware(request: Request, response: HTTPResponse):
    """
    Adds a header to all responses that prevents cross site scripting.

    :param request: Sanic request parameter.

    :param response: Sanic http response parameter.
    """
    response.headers['x-xss-protection'] = '1; mode=block'


def https_redirect_middleware(request: Request):
    """
    Redirects all http requests to https.

    :param request: Sanic request parameter.

    :return: redirect_url
    """
    if request.url.startswith('http://'):
        url = request.url.replace('http://', 'https://', 1)
        return redirect(url)


def hash_pw(password: str):
    """
    Turns passed text into hashed password

    :param password: Password to be hashed.

    :return: hashed_password
    """
    return hashlib.pbkdf2_hmac('sha512', password.encode('utf-8'), config['AUTH']['SECRET'].encode('utf-8'), 100000)


def get_ip(request: Request):
    """
    Retrieves ip address from request.

    :param request: Sanic request parameter.

    :return: ip
    """
    return request.remote_addr if request.remote_addr else request.ip


def path_exists(path):
    """
    Checks if path exists and isn't empty, and creates it if it doesn't.

    :param path: Path being checked.

    :return: exists
    """
    exists = os.path.exists(path)
    if not exists:
        os.makedirs(path)
    return exists and os.listdir(path)



from sanic import Sanic
from sanic.exceptions import ServerError
from sanic.response import json as sanic_json
from sanic.response import text, file

from sanic_security.core.authentication import register, login, requires_authentication, logout
from sanic_security.core.authorization import require_permissions, require_roles
from sanic_security.core.initializer import initialize_security
from sanic_security.core.models import SecurityError, Permission, Role, VerificationSession, CaptchaSession
from sanic_security.core.recovery import attempt_recovery, fulfill_recovery_attempt
from sanic_security.core.utils import xss_prevention_middleware
from sanic_security.core.verification import requires_captcha, request_captcha, requires_verification, verify_account, \
    request_verification
from sanic_security.lib.ip2proxy import detect_proxy, proxy_detection_middleware

app = Sanic('Sanic Security test server')


def json(message, content, status_code=200):
    payload = {
        'message': message,
        'status_code': status_code,
        'content': content
    }
    return sanic_json(payload, status=status_code)


def check_for_empty(form, *args):
    for key, value in form.items():
        if value is not None:
            if not isinstance(value[0], bool) and not value[0] and key not in args:
                raise ServerError(key + " is empty!", 400)


@app.middleware('response')
async def xxs_middleware(request, response):
    """
    Response middleware test.
    """
    xss_prevention_middleware(request, response)


@app.middleware('request')
async def ip2proxy_middleware(request):
    """
    Request middleware test.
    """
    pass
    #await proxy_detection_middleware(request)


@app.post('api/test/register')
async def on_register(request):
    """
    Registration test without verification or captcha requirements.
    """
    account = await register(request, verified=True)
    return json('Registration Successful!', account.json())


@app.post('api/test/register/verification')
@requires_captcha()
async def on_register_verification(request, captcha_session):
    """
    Registration test with all built-in requirements.
    """
    verification_session = await register(request)
    await verification_session.text_code()
    response = json('Registration successful', verification_session.account.json())
    verification_session.encode(response, secure=False)
    return response


@app.post('api/test/register/verify')
@requires_verification()
async def on_verify(request, verification_session):
    """
    Attempt to verify account and allow access if unverified.
    """
    await verify_account(verification_session)
    return json('Verification successful!', verification_session.json())


@app.get('api/test/captcha/img')
async def on_captcha_img(request):
    """
    Retrieves captcha image from captcha session.
    """
    img_path = await CaptchaSession.captcha_img(request)
    return await file(img_path)


@app.get('api/test/captcha')
async def on_request_captcha(request):
    """
    Requests captcha session for client.
    """
    captcha_session = await request_captcha(request)
    response = json('Captcha request successful!', captcha_session.json())
    captcha_session.encode(response, secure=False)
    return response


@app.post('api/test/verification/resend')
async def resend_verification_request(request):
    """
    Resends verification code if somehow lost.
    """
    verification_session = await VerificationSession().decode(request)
    await verification_session.text_code()
    return json('Verification code resend successful', verification_session.json())


@app.post('api/test/verification/request')
async def new_verification_request(request):
    """
    Creates new verification code.
    """

    verification_session = await request_verification(request)
    await verification_session.text_code()
    response = json('Verification request successful', verification_session.json())
    verification_session.encode(response, secure=False)
    return response


@app.post('api/test/login')
async def on_login(request):
    """
    User login, creates and encodes authentication session.
    """
    authentication_session = await login(request)
    response = json('Login successful!', authentication_session.account.json())
    authentication_session.encode(response, secure=False)
    return response


@app.post('api/test/logout')
async def on_logout(request):
    """
    User logout, invalidates client authentication session.
    """
    authentication_session = await logout(request)
    response = json('Logout successful!', authentication_session.account.json())
    return response


@app.post('api/test/role/admin')
@requires_authentication()
async def on_create_admin(request, authentication_session):
    """
    Creates 'Admin' and 'Mod' roles to be used for testing role based authorization.
    """
    client = authentication_session.account
    await Role().create(account=client, name='Admin')
    await Role().create(account=client, name='Mod')
    return json('Roles added to your account!', client.json())


@app.post('api/test/perms/admin')
@requires_authentication()
async def on_create_admin_perm(request, authentication_session):
    """
    Creates 'admin:update' and 'admin:add' permissions to be used for testing wildcard based authorization.
    """
    client = authentication_session.account
    await Permission().create(account=client, wildcard='admin:update', decription="")
    await Permission().create(account=client, wildcard='admin:add')
    return json('Permissions added to your account!', client.json())


@app.get('api/test/client')
@detect_proxy()
@requires_authentication()
async def on_test_client(request, authentication_session):
    """
    Retrieves authenticated client username.
    """
    return text('Hello ' + authentication_session.account.username + '!')


@app.get('api/test/perm')
@require_permissions('admin:update')
async def on_test_perm(request, authentication_session):
    """
    Tests client wildcard permissions authorization access.
    """
    return text('Admin who can only update gained access!')


@app.get('api/test/role')
@require_roles('Admin', 'Mod')
async def on_test_role(request, authentication_session):
    """
    Tests client role authorization access.
    """
    return text('Admin gained access!')


@app.post('api/test/recovery/attempt')
@requires_captcha()
async def on_recovery_attempt(request, captcha_session):
    """
    Attempts to recover account via changing password, requests verification to ensure the recovery attempt was made
    by account owner.
    """
    verification_session = await attempt_recovery(request)
    await verification_session.text_code()
    response = json('A recovery attempt has been made, please verify account ownership.', verification_session.json())
    verification_session.encode(response, secure=False)
    return response


@app.post('api/test/recovery/fulfill')
@requires_verification()
async def on_recovery_fulfill(request, verification_session):
    """
    Changes and recovers an account's password once recovery attempt was determined to have been made by account owner.
    """
    await fulfill_recovery_attempt(request, verification_session)
    return json('Account recovered successfully', verification_session.account.json())


@app.exception(SecurityError)
async def on_error(request, exception):
    return json('An error has occurred!', {
        'error': type(exception).__name__,
        'summary': str(exception)
    }, status_code=exception.status_code)


if __name__ == '__main__':
    initialize_security(app)
    app.run(host='0.0.0.0', port=8000, debug=True, workers=4)



from sanic.request import Request

from sanic_security.core.authentication import account_error_factory
from sanic_security.core.models import AuthenticationSession, Account, VerificationSession
from sanic_security.core.utils import hash_pw
from sanic_security.core.verification import request_verification


async def fulfill_recovery_attempt(request: Request, verification_session: VerificationSession):
    """
    Recovers an account by setting the password to a new one once recovery attempt was determined to be made
    by the account owner.

    :param request: Sanic request parameter. All request bodies are sent as form-data with the following arguments:
    password.

    :param verification_session: Verification session containing account being recovered.
    """
    verification_session.account.password = hash_pw(request.form.get('password'))
    await AuthenticationSession.filter(account=verification_session.account, valid=True,
                                       deleted=False).update(valid=False)
    await verification_session.account.save(update_fields=['password'])


async def attempt_recovery(request: Request):
    """
    Requests a verification session to ensure that the recovery attempt was made by the account owner.

    :param request: Sanic request parameter. This request is sent with the following url argument: email.

    return: verification_session
    """

    account = await Account.filter(email=request.form.get('email')).first()
    account_error_factory.throw(account)
    return await request_verification(request, account)



import functools
from fnmatch import fnmatch

from sanic.request import Request

from sanic_security.core.authentication import authenticate
from sanic_security.core.models import Role, Permission


async def check_permissions(request: Request, *required_permissions: str):
    """
    Checks if the client has the required permissions.

    :param request: Sanic request parameter.

    :param required_permissions: The permissions required to authorize an action.

    :raises InsufficientPermissionError:
    """

    authentication_session = await authenticate(request)
    client_permissions = await Permission.filter(account=authentication_session.account).all()
    for required_permission in required_permissions:
        for client_permission in client_permissions:
            if fnmatch(required_permission, client_permission.wildcard):
                break
        else:
            raise Permission.InsufficientPermissionError()
    return authentication_session


async def check_roles(request: Request, *required_roles: str):
    """
    Checks if the client has the required roles.

    :param request: Sanic request parameter.

    :param required_roles: The roles required to authorize an action.

    :raises InsufficientRoleError:
    """

    authentication_session = await authenticate(request)
    for role in required_roles:
        if await Role.filter(account=authentication_session.account, name=role).exists():
            break
    else:
        raise Role.InsufficientRoleError()
    return authentication_session


def require_permissions(*required_permissions: str):
    """
    Checks if the client has the required permissions.

    :param required_permissions: The permissions required to authorize an action.

    :raises InsufficientPermissionError:
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            authentication_session = await check_permissions(request, *required_permissions)
            return await func(request, authentication_session, *args, **kwargs)

        return wrapped

    return wrapper


def require_roles(*required_roles: str):
    """
    Checks if the client has the required roles.

    :param required_roles: The roles required to authorize an action.

    :raises InsufficientRoleError:
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            authentication_session = await check_roles(request, *required_roles)
            return await func(request, authentication_session, *args, **kwargs)

        return wrapped

    return wrapper



from sanic import Sanic

from sanic_security.core.config import config
from sanic_security.core.models import Session
from sanic_security.lib.ip2proxy import initialize_ip2proxy
from sanic_security.lib.tortoise import initialize_tortoise


def initialize_security(app: Sanic):
    """
    Initializes sanic-security.

    :param app: Sanic object used to add tasks too.
    """
    initialize_tortoise(app)
    Session.initialize_cache(app)
    if config.has_section('IP2PROXY'):
        initialize_ip2proxy(app)




import functools
import re

from sanic.request import Request
from tortoise.exceptions import IntegrityError, ValidationError

from sanic_security.core.models import Account, SessionFactory, AuthenticationSession, Session
from sanic_security.core.utils import hash_pw
from sanic_security.core.verification import request_verification

session_factory = SessionFactory()
account_error_factory = Account.ErrorFactory()
session_error_factory = Session.ErrorFactory()


async def register(request: Request, verified: bool = False, disabled: bool = False):
    """
    Creates a new account. This is the recommend and most secure method for registering accounts' with sanic-security.

    :param request: Sanic request parameter. All request bodies are sent as form-data with the following arguments:
    email, username, password, phone.

    :param verified: If false, account being registered must be verified before use.

    :param disabled: If true, account being registered must be enabled before use.

    :raises AccountError:

    :return: account if verified or verification_session if not verified
    """
    forms = request.form
    if not re.search('[^@]+@[^@]+.[^@]+', forms.get('email')):
        raise Account.InvalidEmailError()
    try:
        account = await Account.create(email=forms.get('email'), username=forms.get('username'),
                                       password=hash_pw(forms.get('password')), phone=forms.get('phone'),
                                       verified=verified, disabled=disabled)
        return await request_verification(request, account) if not verified else account
    except IntegrityError:
        raise Account.ExistsError()
    except ValidationError:
        raise Account.TooManyCharsError()


async def login(request: Request):
    """
    Creates an authentication session that is used to verify the account making requests requiring authentication.

    :param request: Sanic request parameter. All request bodies are sent as form-data with the following arguments:
    email, password.

    :raises AccountError:

    :return: authentication_session
    """
    form = request.form
    account = await Account.filter(email=form.get('email')).first()
    account_error_factory.throw(account)
    if account.password == hash_pw(form.get('password')):
        authentication_session = await session_factory.get('authentication', request, account=account)
        return authentication_session
    else:
        raise Account.IncorrectPasswordError()


async def logout(request: Request):
    """
    Invalidates client's authentication session.

    :param request: Sanic request parameter.
    """
    authentication_session = await AuthenticationSession().decode(request)
    authentication_session.valid = False
    await authentication_session.save(update_fields=['valid'])
    return authentication_session


async def authenticate(request: Request):
    """
    Authenticates the client's current authentication session.

    :raises AccountError:

    :raises SessionError:

    :return: authentication_session
    """
    authentication_session = await AuthenticationSession().decode(request)
    session_error_factory.throw(authentication_session)
    await authentication_session.crosscheck_location(request)
    account_error_factory.throw(authentication_session.account)
    return authentication_session


def requires_authentication():
    """
    Authenticates the client's current authentication session.

    :raises AccountError:

    :raises SessionError:

    :return: func(request, authentication_session, *args, **kwargs)
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            authentication_session = await authenticate(request)
            return await func(request, authentication_session, *args, **kwargs)

        return wrapped

    return wrapper



import asyncio
import functools
import shutil

import aioIP2Proxy
import aiofiles
import aiohttp
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from sanic.request import Request

from sanic_security.core.config import config
from sanic_security.core.models import SecurityError
from sanic_security.core.utils import path_exists, get_ip

ip2proxy_database = aioIP2Proxy.IP2Proxy()


class IP2ProxyError(SecurityError):
    pass


class ProxyDetectedError(IP2ProxyError):
    def __init__(self):
        super(ProxyDetectedError, self).__init__('An attempt was made to access a resource utilizing a forbidden '
                                                 'proxy.', 403)


async def cache_ip2proxy_database():
    """
    Caches a new IP2Proxy database.
    """
    code = config['IP2PROXY']['code']
    key = config['IP2PROXY']['key']
    loop = asyncio.get_running_loop()
    async with aiohttp.ClientSession() as session:
        url = "https://www.ip2location.com/download/?token={0}&file={1}".format(key, code)
        async with session.get(url) as resp:
            zip_path = './resources/security-cache/ip2proxy/ip2proxy.zip'
            async with aiofiles.open(zip_path, mode="wb") as f:
                try:
                    await f.write(await resp.read())
                    await loop.run_in_executor(None, shutil.unpack_archive, zip_path,
                                               './resources/security-cache/ip2proxy')
                except shutil.ReadError:
                    raise IP2ProxyError('Unzipping has failed due to the download rate limit or incorrect credentials.')


def initialize_ip2proxy(app):
    """
    Initializes a async cron job that runs every 00:15 GMT to refresh the IP2Proxy database.
    """
    scheduler = AsyncIOScheduler()

    @app.listener("before_server_start")
    async def init_ip2proxy_cron(app, loop):
        if not path_exists('./resources/security-cache/ip2proxy'):
            await cache_ip2proxy_database()
        scheduler.add_job(cache_ip2proxy_database, 'cron', minute='15', hour='0', month='*', week='*', day='*')
        scheduler.start()

    @app.listener("after_server_stop")
    async def shutdown_ip2proxy_cron(app, loop):
        scheduler.shutdown()


async def proxy_detection(ip: str):
    """
    Reads local database file and crosschecks passed ip address to determine if it is a known proxy.

    :param ip: Ip address being crosschecked.

    """
    await ip2proxy_database.open('./resources/security-cache/ip2proxy/' + config['IP2PROXY']['bin'])
    if await ip2proxy_database.is_proxy(ip) > 0:
        raise ProxyDetectedError()
    await ip2proxy_database.close()


def detect_proxy():
    """
    Reads local database file and crosschecks passed ip address to determine if it is a known proxy.

    :raises AccountError:

    :raises SessionError:

    :return: func(request, authentication_session, *args, **kwargs)
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            await proxy_detection_middleware(request)
            return await func(request, *args, **kwargs)

        return wrapped

    return wrapper


async def proxy_detection_middleware(request: Request):
    """
    Reads local database file and crosschecks passed ip address to determine if it is a known proxy.

    :param request: Sanic request parameter.
    """
    await proxy_detection(get_ip(request))



import functools

from sanic.request import Request

from sanic_security.core.models import Account, VerificationSession, CaptchaSession, SessionFactory, Session

session_factory = SessionFactory()
session_error_factory = Session.ErrorFactory()


async def request_captcha(request: Request):
    """
    Creates a captcha session associated with an account.

    :param request: Sanic request parameter.

    :return: captcha_session
    """
    return await session_factory.get('captcha', request)


async def captcha(request: Request):
    """
    Enforces captcha.

    :param request: Sanic request parameter. All request bodies are sent as form-data with the following arguments:
    captcha.

    :raises SessionError:

    :return: captcha_session
    """
    captcha_session = await CaptchaSession().decode(request)
    session_error_factory.throw(captcha_session)
    await captcha_session.crosscheck_code(request.form.get('captcha'))
    return captcha_session


def requires_captcha():
    """
    Enforced captcha.

    :raises SessionError:

    :return: await func(request, captcha_session, *args, **kwargs)
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            captcha_session = await captcha(request)
            return await func(request, captcha_session, *args, **kwargs)

        return wrapped

    return wrapper


async def request_verification(request: Request, account: Account = None):
    """
    Creates a verification session associated with an account.

    :param request: Sanic request parameter.

    :param account: The account being associated with the verification session.

    :return: verification_session
    """
    if account is None:
        verification_session = await VerificationSession().decode(request)
        account = verification_session.account
    return await session_factory.get('verification', request, account=account)


async def verify(request: Request):
    """
    Enforces verification and validates attempts.

    :param request: Sanic request parameter. All request bodies are sent as form-data with the following arguments:
    code.

    :raises SessionError:

    :return: verification_session
    """
    verification_session = await VerificationSession().decode(request)
    session_error_factory.throw(verification_session)
    await verification_session.crosscheck_code(request.form.get('code'))
    return verification_session


async def verify_account(verification_session: VerificationSession):
    """
    Verifies account associated to a verification session.

    :param verification_session: Verification session containing account being verified.
    """
    verification_session.account.verified = True
    await verification_session.account.save(update_fields=['verified'])
    return verification_session


def requires_verification():
    """
    Enforces verification and validates attempts.

    :raises SessionError:

    :return: func(request, verification_session, *args, **kwargs)
    """

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(request, *args, **kwargs):
            verification_session = await verify(request)
            return await func(request, verification_session, *args, **kwargs)

        return wrapped

    return wrapper



import asyncio
import datetime
import random
import string
import uuid
import aiofiles
import jwt
from captcha.image import ImageCaptcha
from jwt import DecodeError
from sanic import Sanic
from sanic.exceptions import ServerError
from sanic.request import Request
from sanic.response import HTTPResponse
from tortoise import fields, Model
from sanic_security.core.config import config
from sanic_security.core.utils import path_exists, get_ip
from sanic_security.lib.smtp import send_email
from sanic_security.lib.twilio import send_sms


class BaseErrorFactory:
    """
    Easily raise or retrieve errors based off of variable values. Validates the ability for a model to be utilized.
    """

    def get(self, model):
        """
        Retrieves an error if certain conditions are met.
        :return: error
        """
        raise NotImplementedError()

    def throw(self, model):
        """
        Retrieves an error and raises it if certain conditions are met.
        :return: error
        """
        error = self.get(model)
        if error:
            raise error


class SecurityError(ServerError):
    """
    Base error for all Sanic Security related errors.
    """

    def __init__(self, message, code=None):
        super().__init__(message, code)


class BaseModel(Model):
    """
    Base Sanic Security model that all other models derive from.
    """

    id = fields.IntField(pk=True)
    account = fields.ForeignKeyField('models.Account', null=True)
    uid = fields.UUIDField(unique=True, default=uuid.uuid1, max_length=36)
    date_created = fields.DatetimeField(auto_now_add=True)
    date_updated = fields.DatetimeField(auto_now=True)
    deleted = fields.BooleanField(default=False)

    def json(self):
        raise NotImplementedError()

    class Meta:
        abstract = True

    class NotFoundError(SecurityError):
        def __init__(self, message):
            super().__init__(message, 404)

    class DeletedError(SecurityError):
        def __init__(self, message):
            super().__init__(message, 404)


class Account(BaseModel):
    """
    Contains all identifiable user information such as username, email, and more. All passwords must be hashed when
    being created in the database using the hash_pw(str) method.
    """
    username = fields.CharField(max_length=45)
    email = fields.CharField(unique=True, max_length=45)
    phone = fields.CharField(unique=True, max_length=20, null=True)
    password = fields.BinaryField()
    disabled = fields.BooleanField(default=False)
    verified = fields.BooleanField(default=False)

    class ErrorFactory(BaseErrorFactory):
        def get(self, model):
            error = None
            if not model:
                error = Account.NotFoundError('This account does not exist.')
            elif model.deleted:
                error = Account.DeletedError('This account has been permanently deleted.')
            elif model.disabled:
                error = Account.DisabledError()
            elif not model.verified:
                error = Account.UnverifiedError()
            return error

    def json(self):
        return {
            'uid': str(self.uid),
            'date_created': str(self.date_created),
            'date_updated': str(self.date_updated),
            'email': self.email,
            'username': self.username,
            'disabled': self.disabled,
            'verified': self.verified
        }

    class AccountError(SecurityError):
        def __init__(self, message, code):
            super().__init__(message, code)

    class ExistsError(AccountError):
        def __init__(self):
            super().__init__('Account with this email or phone number already exists.', 409)

    class TooManyCharsError(AccountError):
        def __init__(self):
            super().__init__('Email, username, or phone number is too long.', 400)

    class InvalidEmailError(AccountError):
        def __init__(self):
            super().__init__('Please use a valid email format such as you@mail.com.', 400)

    class DisabledError(AccountError):
        def __init__(self):
            super().__init__("This account has been disabled.", 401)

    class IncorrectPasswordError(AccountError):
        def __init__(self):
            super().__init__('The password provided is incorrect.', 401)

    class UnverifiedError(AccountError):
        def __init__(self):
            super().__init__('Account requires verification.', 401)


class Session(BaseModel):
    """
    Used specifically for client side tracking. For example, an authentication session is stored on the client's browser
    in order to identify the client. All sessions should be created using the SessionFactory().
    """

    expiration_date = fields.DatetimeField(null=True)
    valid = fields.BooleanField(default=True)
    ip = fields.CharField(max_length=16)
    attempts = fields.IntField(default=0)
    code = fields.CharField(max_length=8, null=True)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.cookie = config['AUTH']['name'].strip() + '_' + self.__class__.__name__

    def json(self):
        return {
            'uid': str(self.uid),
            'date_created': str(self.date_created),
            'date_updated': str(self.date_updated),
            'expiration_date': str(self.expiration_date),
            'account': self.account.email if isinstance(self.account, Account) else None,
            'valid': self.valid,
            'attempts': self.attempts
        }

    @staticmethod
    def initialize_cache(app: Sanic):
        """
        Caches up to 100 code and image variations.
        """

        @app.listener("before_server_start")
        async def generate_codes(app, loop):
            session_cache = './resources/security-cache/session/'
            loop = asyncio.get_running_loop()
            image = ImageCaptcha(190, 90, fonts=[config['AUTH']['captcha_font']])
            if not path_exists(session_cache):
                async with aiofiles.open(session_cache + 'codes.txt', mode="w") as f:
                    for i in range(100):
                        code = ''.join(random.choices(string.ascii_letters + string.digits, k=8))
                        await f.write(code + ' ')
                        await loop.run_in_executor(None, image.write, code[:6], session_cache + code[:6] + '.png')

    @staticmethod
    async def get_code():
        """
        Retrieves a random cached code from a codes.txt file

        :return: code
        """
        async with aiofiles.open('./resources/security-cache/session/codes.txt', mode='r') as f:
            codes = await f.read()
            return random.choice(codes.split())

    async def crosscheck_code(self, code: str):
        """
        Used to check if code passed is equivalent to the session code.

        :param code: Code being cross-checked with session code.
        """
        if self.attempts >= 5:
            raise self.MaximumAttemptsError
        elif self.code != code:
            self.attempts += 1
            await self.save(update_fields=['attempts'])
            raise self.CrosscheckError()
        else:
            self.valid = False
            await self.save(update_fields=['valid'])

    def encode(self, response: HTTPResponse, secure=True, same_site: str = 'lax'):
        """
        Transforms session into jwt and then is stored in a cookie.

        :param response: Response used to store cookie.

        :param secure: If true, connections must be SSL encrypted (aka https).

        :param same_site: Allows you to declare if your cookie should be restricted to a first-party or same-site context.
        """

        payload = {
            'date_created': str(self.date_created),
            'uid': str(self.uid),
            'ip': self.ip
        }
        encoded = jwt.encode(payload, config['AUTH']['secret'], 'HS256')
        response.cookies[self.cookie] = encoded
        response.cookies[self.cookie]['secure'] = secure
        response.cookies[self.cookie]['samesite'] = same_site

    def decode_raw(self, request: Request):
        """
        Decodes JWT token in cookie to dict.

        :param request: Sanic request parameter.

        :return: raw
        """
        cookie = request.cookies.get(self.cookie)
        try:
            if not cookie:
                raise DecodeError('Token can not be null.')
            else:
                return jwt.decode(cookie, config['AUTH']['secret'], 'HS256')
        except DecodeError as e:
            raise Session.DecodeError(e)

    async def decode(self, request: Request):
        """
        Decodes JWT token in cookie and transforms into session.

        :param request: Sanic request parameter.

        :return: session
        """

        decoded = self.decode_raw(request)
        return await self.filter(uid=decoded.get('uid')).prefetch_related('account').first()

    class Meta:
        abstract = True

    class ErrorFactory(BaseErrorFactory):
        def get(self, model):
            error = None

            if model is None:
                error = Session.NotFoundError('Session could not be found.')
            elif not model.valid:
                error = Session.InvalidError()
            elif model.deleted:
                error = Session.DeletedError('Session has been deleted.')
            elif datetime.datetime.now(datetime.timezone.utc) >= model.expiration_date:
                error = Session.ExpiredError()
            return error

    class SessionError(SecurityError):
        def __init__(self, message, code):
            super().__init__(message, code)

    class MaximumAttemptsError(SessionError):
        def __init__(self):
            super().__init__('You\'ve reached the maximum amount of attempts for this session.', 401)

    class DecodeError(SessionError):
        def __init__(self, exception):
            super().__init__('Session cookie could not be decoded. ' + str(exception), 400)

    class InvalidError(SessionError):
        def __init__(self):
            super().__init__('Session is invalid.', 401)

    class CrosscheckError(SessionError):
        def __init__(self):
            super().__init__('Session crosschecking attempt was incorrect', 401)

    class ExpiredError(SessionError):
        def __init__(self):
            super().__init__('Session has expired', 401)


class SessionFactory:
    """
    Prevents human error when creating sessions.
    """

    def generate_expiration_date(self, days: int = 0, minutes: int = 0):
        """
        Creates an expiration date. Adds days to current datetime.

        :param days: days to be added to current time.

        :param minutes: minutes to be added to the current time.

        :return: expiration_date
        """
        return datetime.datetime.utcnow() + datetime.timedelta(days=days, minutes=minutes)

    async def _account_via_decoded(self, request: Request, session: Session):
        """
        Extracts account from decoded session. This method was created purely to prevent repetitive code.

        :param request: Sanic request parameter.

        :param session: Session being decoded to retrieve account from

        :return: account
        """
        decoded_session = await session.decode(request)
        return decoded_session.account

    async def get(self, session_type: str, request: Request, account: Account = None):
        """
        Creates and returns a session with all of the fulfilled requirements.

        :param session_type: The type of session being retrieved. Available types are: captcha, verification, and
        authentication.

        :param request: Sanic request parameter.

        :param account: Account being associated to to a session.

        :return: session
        """
        code = await Session.get_code()
        if session_type == 'captcha':
            return await CaptchaSession.create(ip=get_ip(request), code=code[:6],
                                               expiration_date=self.generate_expiration_date(minutes=1))
        elif session_type == 'verification':
            return await VerificationSession.create(code=code, ip=get_ip(request), account=account,
                                                    expiration_date=self.generate_expiration_date(minutes=5))
        elif session_type == 'authentication':
            return await AuthenticationSession.create(account=account, ip=get_ip(request),
                                                      expiration_date=self.generate_expiration_date(days=30))
        else:
            raise ValueError('Invalid session type.')


class VerificationSession(Session):
    """
    Used to verify an account's email or mobile. Can be used in order to validate the person utilizing
    an account is the actual owner.
    """

    async def text_code(self, code_prefix="Your code is: "):
        """
        Sends verification code via text.

        :param code_prefix: Message being sent with code, for example "Your code is: ".
        """
        await send_sms(self.account.phone, code_prefix + self.code)

    async def email_code(self, subject="Session Code", code_prefix='Your code is:\n\n '):
        """
        Sends verification code via email.

        :param code_prefix: Message being sent with code, for example "Your code is: ".

        :param subject: Subject of email being sent with code.
        """
        await send_email(self.account.email, subject, code_prefix + self.code)


class CaptchaSession(Session):
    """
    Validates a client as human by correctly entering a captcha challenge.
    """

    @staticmethod
    async def captcha_img(request):
        """
        Retrieves image path of captcha.

        :return: captcha_img_path
        """
        decoded_captcha = await CaptchaSession().decode(request)
        return './resources/security-cache/session/' + decoded_captcha.code + '.png'


class AuthenticationSession(Session):
    class UnknownLocationError(Session.SessionError):
        def __init__(self):
            super().__init__('Session in an unknown location.', 401)

    async def crosscheck_location(self, request):
        """
        Checks if client using session is in a known location (ip address). Prevents cookie jacking.

        :raises UnknownLocationError:
        """

        if not await AuthenticationSession.filter(ip=get_ip(request), account=self.account).exists():
            raise AuthenticationSession.UnknownLocationError()


class Role(BaseModel):
    """
    Assigned to an account to authorize an action. Used for role based authorization.
    """
    name = fields.CharField(max_length=45)

    def json(self):
        return {
            'uid': str(self.uid),
            'date_created': str(self.date_created),
            'date_updated': str(self.date_updated),
            'name': self.name,
        }

    class InsufficientRoleError(SecurityError):
        def __init__(self):
            super().__init__('Insufficient roles required for this action.', 403)


class Permission(BaseModel):
    """
    Assigned to an account to authorize an action. Used for wildcard based authorization.
    """
    wildcard = fields.CharField(max_length=45)

    def json(self):
        return {
            'uid': str(self.uid),
            'date_created': str(self.date_created),
            'date_updated': str(self.date_updated),
            'wildcard': self.wildcard,
        }

    class InsufficientPermissionError(SecurityError):
        def __init__(self):
            super().__init__('Insufficient permissions required for this action.', 403)



import aiohttp

from sanic_security.core.config import config


async def send_sms(to, msg):
    """
    Sends a text directly to a phone number using Twilio.

    :param to: The phone number receiving a text, for example: +12092819472.
    :param msg: The message to be sent as a text.
    """
    account_sid = config['TWILIO']['sid']
    auth_token = config['TWILIO']['token']
    from_num = '+' + config['TWILIO']['from']
    if account_sid and auth_token and from_num:
        auth = aiohttp.BasicAuth(login=account_sid, password=auth_token)
        async with aiohttp.ClientSession(auth=auth) as session:
            await session.post(f'https://api.twilio.com/2010-04-01/Accounts/{account_sid}/Messages.json',
                               data={'From': from_num, 'To': '+' + to, 'Body': msg})
    else:
        raise RuntimeWarning('Twilio credentials not found.')



from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from aiosmtplib import send
from sanic_security.core.config import config


async def send_email(to, subj, msg, text_type="plain"):
    """
    Sends an email using SMTP.

    :param to: To email.
    :param subj: Email subject.
    :param msg: Email message.
    :param text_type: Can be html or plain.
    """
    message = MIMEMultipart("alternative")
    message["From"] = config['SMTP']['from']
    message["To"] = to
    message["Subject"] = subj
    mime_text = MIMEText(msg, text_type, "utf-8")
    message.attach(mime_text)
    await send(
        message,
        hostname=config['SMTP']['host'],
        port=int(config['SMTP']['port']),
        username=config['SMTP']['username'],
        password=config['SMTP']['password'],
        use_tls=config['SMTP']['tls'] == 'true',
        start_tls=config['SMTP']['start_tls'] == 'true'
    )



from sanic import Sanic
from tortoise import Tortoise
from tortoise.contrib.sanic import register_tortoise

from sanic_security.core.config import config


def initialize_tortoise(app: Sanic):
    """
    Initializes tortoise-orm.
    """

    @app.listener("before_server_start")
    async def init_orm(app, loop):
        username = config['TORTOISE']['username']
        password = config['TORTOISE']['password']
        endpoint = config['TORTOISE']['endpoint']
        schema = config['TORTOISE']['schema']
        engine = config['TORTOISE']['engine']
        models = config['TORTOISE']['models'].replace(' ', '').split(',')
        url = engine + '://{0}:{1}@{2}/{3}'.format(username, password, endpoint, schema)
        await Tortoise.init(db_url=url, modules={"models": models})
        if config['TORTOISE']['generate'] == 'true':
            await Tortoise.generate_schemas()

    @app.listener("after_server_stop")
    async def close_orm(app, loop):
        await Tortoise.close_connections()


